/*  gngeo a neogeo emulator
 *  Copyright (C) 2001 Peponas Mathieu
 * 
 *  This program is free software; you can redistribute it and/or modify  
 *  it under the terms of the GNU General Public License as published by   
 *  the Free Software Foundation; either version 2 of the License, or    
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "SDL.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#include "unzip.h"
#include "memory.h"
#include "video.h"
#include "emu.h"
#include "fileio.h"
#include "neocrypt.h"
#include "screen.h"
#include "conf.h"
#include "pbar.h"
#include "driver.h"
#include "sound.h"

#ifdef USE_GUI
#include "gui/gui.h"
#endif

unsigned char *current_buf;
//extern unsigned char fix_buffer[0x20000];
char *rom_file;
#if 0
typedef struct _game_list {
    char game[10];
    struct _game_list *next;
} GAME_LIST;
GAME_LIST *gl;

static SDL_bool list_game_add(char *game)
{
    GAME_LIST *t = gl;
    if (t != NULL) {
	while (t->next != NULL) {
	    if (strncmp(t->game, game, strlen(game)) == 0
		&& strlen(game) == strlen(t->game)) {
		return SDL_FALSE;
	    }
	    t = t->next;
	}
	t->next = (GAME_LIST *) malloc(sizeof(GAME_LIST));
	strcpy(t->next->game, game);
	t->next->next = NULL;
	return SDL_TRUE;
    } else {
	gl = (GAME_LIST *) malloc(sizeof(GAME_LIST));
	strcpy(gl->game, game);
	gl->next = NULL;
	return SDL_TRUE;
    }
}

void list_game(void)
{
    char buf[512];
    FILE *f;
    char game[10], type[5];
    char *t;
    char *romrc_path=CF_STR(cf_get_item_by_name("romrc"));
    if ((f = fopen(romrc_path, "r")) == NULL) {
	printf("Couldn't find %s\n", romrc_path);
	exit(1);
    }
    while (!feof(f)) {
	fgets(buf, 510, f);
	if (sscanf(buf, "game %s %s\n", game, type) >= 2) {
	    if (list_game_add(game) == SDL_TRUE) {
		t = strchr(buf, '"');
		printf("%8s  %s", game, (t == NULL ? "\n" : t));
	    }
	}
    }
    fclose(f);
}
#endif

#if 0
char *get_next_game(FILE * f)
{
    static char buf[512];
    static char game[10], type[5];

    while (!feof(f)) {
	fgets(buf, 510, f);
	if (sscanf(buf, "game %s %s\n", game, type) >= 2) {
	    if (strcmp(type, "MGD2") == 0)
		conf.rom_type = MGD2;
	    else if (strcmp(type, "MVS_CMC42") == 0)
		conf.rom_type = MVS_CMC42;
	    else if (strcmp(type, "MVS_CMC50") == 0)
		conf.rom_type = MVS_CMC50;
	    else
		conf.rom_type = MVS;
	    return game;
	}
    }
    return NULL;
}

int search_game(FILE * f, char *name)
{
    static char buf[512];
    static char game[10], type[5];

    while (!feof(f)) {
	fgets(buf, 510, f);
	if (sscanf(buf, "game %s %s\n", game, type) >= 2) {
	    if (strlen(game) != strlen(name))
		continue;
	    if (strncmp(game, name, strlen(name)) == 0) {
		if (strcmp(type, "MGD2") == 0)
		    conf.rom_type = MGD2;
		else if (strcmp(type, "MVS_CMC42") == 0)
		    conf.rom_type = MVS_CMC42;
		else if (strcmp(type, "MVS_CMC50") == 0)
		    conf.rom_type = MVS_CMC50;
		else
		    conf.rom_type = MVS;
		return 0;
	    }
	}
    }
    return -1;
}


static int zfread(unzFile * f, void *buffer, int length)
{
    unsigned char *buf = (Uint8*)buffer;
    unsigned char *tempbuf;
    unsigned int totread, r, i;
    int totlength=length;
    totread = 0;
    tempbuf=alloca(4097);

    while (length) {

	r = length;
	if (r > 4096)
	    r = 4096;

	r = unzReadCurrentFile(f, tempbuf, r);
	if (r == 0) {
	    terminate_progress_bar();
	    return totread;
	}
	memcpy(buf, tempbuf, r);

	buf += r;
	totread += r;
	length -= r;
	update_progress_bar(totread,totlength);
    }

    terminate_progress_bar();
    return totread;
}

static int zfread_alternate(unzFile * f, void *buffer, int length, int inc)
{
    unsigned char *buf = buffer;
    unsigned char tempbuf[4096];
    unsigned int totread, r, i;
    int totlength=length;
    totread = 0;
    

    while (length) {

	r = length;
	if (r > 4096)
	    r = 4096;

	r = unzReadCurrentFile(f, tempbuf, r);
	if (r == 0) {
	    terminate_progress_bar();
	    return totread;
	}
	for (i = 0; i < r; i++) {
	    *buf = tempbuf[i];
	    buf += inc;
	}
	totread += r;
	length -= r;
	update_progress_bar(totread,totlength);
    }

    terminate_progress_bar();
    return totread;
}

SDL_bool open_romset(FILE * ini, char *romname,int is_zip_file)
{
    char buf[512], a[64], b[64];
    unsigned int s, size, offset;
    unzFile *zf;
    char *zipfile;
    char *path = CF_STR(cf_get_item_by_name("rompath"));//conf.rom_path;
    int len = strlen(path) + strlen(romname) + 6;

    if (is_zip_file)
	zipfile = romname;
    else {
	zipfile = alloca(len);
	memset(zipfile, 0, len);
	
	strcat(zipfile, path);
	strcat(zipfile, "/");
	strcat(zipfile, romname);
	strcat(zipfile, ".zip");
    }
//    printf("\nzipfile=%s\n",zipfile);
    /* on ouvre le zip */
    if ((zf = unzOpen(zipfile)) == NULL && (is_zip_file==0)) {
	printf("Cant' find %s\n", zipfile);
	return SDL_FALSE;
    }

    while (1) {
	fgets(buf, 510, ini);
	if (strcmp(buf, "END\n") == 0)
	    break;
	sscanf(buf, "%s %x\n", a, &s);
	if (strcmp(a, "CPU") == 0) {
		memory.cpu = (unsigned char *) realloc(memory.cpu, s);
	    current_buf = memory.cpu;
	    memory.cpu_size = s;
	} else if (strcmp(a, "SFIX") == 0) {
	    memory.sfix_game = (unsigned char *) realloc(memory.sfix_game, s);
	    memory.fix_game_usage = (unsigned char *) realloc(memory.fix_game_usage, s >> 5);
	    current_buf = memory.sfix_game;
	    memory.sfix_size = s;
	} else if (strcmp(a, "SM1") == 0) {
	    memory.sm1 = (unsigned char *) realloc(memory.sm1, s);
	    current_buf = memory.sm1;
	    memory.sm1_size = s;
	} else if (strcmp(a, "SOUND1") == 0) {
	    memory.sound1_size = s;
	    memory.sound1 = (unsigned char *) realloc(memory.sound1, s);
	    current_buf = memory.sound1;
	} else if (strcmp(a, "SOUND2") == 0) {
	    memory.sound2 = (unsigned char *) realloc(memory.sound2, s);
	    memory.sound2_size = s;
	    current_buf = memory.sound2;
	} else if (strcmp(a, "GFX") == 0) {
	    memory.gfx = (unsigned char *) realloc(memory.gfx, s);
	    memory.gfx_size = s;
	    current_buf = memory.gfx;
	    memory.pen_usage = (unsigned int *) realloc(memory.pen_usage, (s >> 7) * sizeof(int));
	    memset(memory.pen_usage, 0, (s >> 7) * sizeof(int));
	    memory.nb_of_tiles = s >> 7;
	} else if (strcmp(a, "XOR") == 0) { /* Special Section for Crypted ROM: 
					       the size argument correspond to the extra xor need for
					       decryption */
	    conf.extra_xor = s;
	    break;
	} else if (strcmp(a, "BANKSWITCH") == 0) { /* Special Section for Crypted ROM :
						      the size parameter correspond to the 
						      address of the handler
						      the first line is to unscramble the bank number
						      the rest is the bankoffset table */
	    int b = 0;
	    memory.bksw_handler = s;
	    fgets(buf, 510, ini);
	    sscanf(buf, "%d %d %d %d %d %d\n",
		   (int *) &memory.bksw_unscramble[0],
		   (int *) &memory.bksw_unscramble[1],
		   (int *) &memory.bksw_unscramble[2],
		   (int *) &memory.bksw_unscramble[3],
		   (int *) &memory.bksw_unscramble[4],
		   (int *) &memory.bksw_unscramble[5]);
	    while (1) {
		fgets(buf, 510, ini);
		if (strcmp(buf, "END\n") == 0)
		    break;
		sscanf("%x\n", &memory.bksw_offset[b]);
		b++;
	    }
	    break;
	} else {
	    printf("Unknow section found: %s\n", a);
	    exit(1);
	}


	while (1) {
	    fgets(buf, 510, ini);
	    if (strcmp(buf, "END\n") == 0)
		break;
	    sscanf(buf, "%s %x %x %s\n", a, &offset, &size, b);
	    
	    if (strcmp(b, "NORM") == 0) {
		if (strcmp(a, "-")) {
		    if (unzLocateFile(zf, a, 2) == UNZ_END_OF_LIST_OF_FILE) {
			unzClose(zf);
			return SDL_FALSE;
		    }
		    if (unzOpenCurrentFile(zf) != UNZ_OK) {
			unzClose(zf);
			return SDL_FALSE;
		    }
		}
		create_progress_bar(a);
		zfread(zf, current_buf + offset, size);
	    } else if (strcmp(b, "ALTERNATE") == 0) {
		if (strcmp(a, "-")) {
		    if (unzLocateFile(zf, a, 2) == UNZ_END_OF_LIST_OF_FILE) {
			unzClose(zf);
			return SDL_FALSE;
		    }
		    if (unzOpenCurrentFile(zf) != UNZ_OK) {
			unzClose(zf);
			return SDL_FALSE;
		    }
		}
		create_progress_bar(a);
		zfread_alternate(zf, current_buf + offset, size, 2);
	    }
	}
    }
    unzClose(zf);
    return SDL_TRUE;
}

int open_rom(char *romname)
{
    FILE *ini;
    unzFile *gz;
    unsigned int i;
    int is_zip_file=0;
    char *romrc_path=CF_STR(cf_get_item_by_name("romrc"));
    char *name=NULL;

    if ((ini = fopen(romrc_path, "r")) == NULL) {
	printf("Can't find %s\n", romrc_path);
	exit(1);
    }

    /* test if romname point to a zipfile or is just the name of the game */
    gz=unzOpen(romname);
    if (gz!=NULL)
	is_zip_file=1;
    unzClose(gz);

    /* the romrc file can contain several romset for one game.
       We try to load the first one, if it fail, we try the
       second, etc.

       if romname is a zip, we try every romset until we find the good one
     */
    do {
	if (is_zip_file) {
	    if ((name=get_next_game(ini))==NULL) {
		printf("No valid romset found for %s\n", romname);
		exit(1);
	    }
	} else {
	    if (search_game(ini, romname)) {
		printf("No valid romset found for %s\n", romname);
		exit(1);
	    }
	}
    } while (open_romset(ini, romname,is_zip_file) == SDL_FALSE);
    
    if (is_zip_file) conf.game=strdup(name);
    else conf.game=strdup(romname);

    if (conf.rom_type == MGD2) {
	create_progress_bar("Convert MGD2");
	convert_mgd2_tiles(memory.gfx, memory.gfx_size);
	convert_mgd2_tiles(memory.gfx, memory.gfx_size);
	terminate_progress_bar();
    }
    if (conf.rom_type == MVS_CMC42) {
	create_progress_bar("Decrypt GFX ");
	kof99_neogeo_gfx_decrypt(conf.extra_xor);
	terminate_progress_bar();
    }
    if (conf.rom_type == MVS_CMC50) {
	create_progress_bar("Decrypt GFX ");
	kof2000_neogeo_gfx_decrypt(conf.extra_xor);
	terminate_progress_bar();
    }

    convert_all_char(memory.sfix_game, memory.sfix_size,
		     memory.fix_game_usage);


    if (CF_BOOL(cf_get_item_by_name("convtile"))) {
	create_progress_bar("Convert tile");
	for (i = 0; i < memory.nb_of_tiles; i++) {
	    convert_tile(i);
	    update_progress_bar(i,memory.nb_of_tiles);
	}
	terminate_progress_bar();
    }

    if (memory.sound2 == NULL) {
	memory.sound2 = memory.sound1;
	memory.sound2_size = memory.sound1_size;
    }
    fclose(ini);

    printf("                                                                             \r");
    return 0;
}

#endif

char *file_basename(char *filename) {
    char *t;
    t=strrchr(filename,'/');
    if (t) return t+1;
    return filename;
}

/* check if dir_name exist. Create it if not */
SDL_bool check_dir(char *dir_name)
{
    DIR *d;

    if (!(d = opendir(dir_name)) && (errno == ENOENT)) {
	mkdir(dir_name, 0755);
	return SDL_FALSE;
    }
    return SDL_TRUE;
}

/* return a char* to $HOME/.gngeo/ */
char *get_gngeo_dir(void) {
    static char *filename=NULL;
    int len = strlen(getenv("HOME")) + strlen("/.gngeo/") + 1;
    int i;
    if (!filename) {
	filename=malloc(len*sizeof(char));
//    filename = (char *) malloc(len);
    sprintf(filename,"%s/.gngeo/",getenv("HOME"));
    }
    check_dir(filename);
    //printf("get_gngeo_dir %s\n",filename);
    return filename;
}

void open_nvram(char *name)
{
    char *filename;
    char *gngeo_dir=get_gngeo_dir();
    FILE *f;
    int len =strlen(name) + strlen(gngeo_dir) + 4; /* ".nv\0" => 4 */
    //strlen(name) + strlen(getenv("HOME")) + strlen("/.gngeo/") + 4;
//    printf("Open nvram %s\n",name);
    filename = (char *) alloca(len);
    sprintf(filename,"%s%s.nv",gngeo_dir,name);
    
    if ((f = fopen(filename, "rb")) == 0)
	return;
    fread(memory.sram, 1, 0x10000, f);
    fclose(f);
}

void save_nvram(char *name)
{
    char *filename;
    char *gngeo_dir=get_gngeo_dir();
    FILE *f;
    int len = strlen(name) + strlen(gngeo_dir) + 4; /* ".nv\0" => 4 */
    //strlen(name) + strlen(getenv("HOME")) + strlen("/.gngeo/") + 4;
    int i;
//    printf("Save nvram %s\n",name);
    for (i = 0xffff; i >= 0; i--) {
	if (memory.sram[i] != 0)
	    break;
    }

    filename = (char *) alloca(len);
    sprintf(filename,"%s%s.nv",gngeo_dir,name);

    f = fopen(filename, "wb");
    fwrite(memory.sram, 1, 8192, f);
    fclose(f);
}

void free_game_memory(void) {

    /* clean up memory */
    free(memory.cpu);memory.cpu=NULL;
    
    free(memory.sm1);memory.sm1=NULL;
    free(memory.sfix_game);memory.sfix_game=NULL;
    if (memory.sound1!=memory.sound2) 
	free(memory.sound2);
    memory.sound2=NULL;
    free(memory.sound1);memory.sound1=NULL;
    free(memory.gfx);memory.gfx=NULL;
    free(memory.pen_usage);memory.pen_usage=NULL;

}

SDL_bool init_game(char *rom_name) {
    DRIVER *dr;
    char *drconf,*gpath;

    dr=dr_get_by_name(rom_name);
    if (!dr) {
#ifdef USE_GUI
	gui_error_box(20,80,264,60,
		      "Error!","No valid romset found for\n%s\n",
		      file_basename(rom_name));
#else
	printf("No valid romset found for %s\n",rom_name);
#endif
	return SDL_FALSE;
    }

    if (conf.game!=NULL) {
	save_nvram(conf.game);
	if (conf.sound) {
	    close_sdl_audio();
	    YM2610_sh_stop();
	    streams_sh_stop();
	}
	free_game_memory();
    }


    /* open transpack if need */
    trans_pack_open(CF_STR(cf_get_item_by_name("transpack")));

    //open_rom(rom_name);
    if (dr_load_game(dr,rom_name)==SDL_FALSE) {
#ifdef USE_GUI
	gui_error_box(20,80,264,60,
		      "Error!","Couldn't load\n%s\n",
		      file_basename(rom_name));
#else
	printf("Can't load %s\n",rom_name);
#endif
	return SDL_FALSE;
    }

    /* per game config */
    gpath=get_gngeo_dir();
    drconf=alloca(strlen(gpath)+strlen(dr->name)+strlen(".cf")+1);
    sprintf(drconf,"%s%s.cf",gpath,dr->name);
    cf_open_file(drconf);

    open_bios();
    open_nvram(conf.game);
    

    sdl_set_title(conf.game);
    init_neo(conf.game);
    if (conf.sound) 
	init_sdl_audio();
    return SDL_TRUE;
}

void free_bios_memory(void) {
    free(memory.ram);memory.ram=NULL;
    free(memory.bios);memory.bios=NULL;
    free(memory.ng_lo);memory.ng_lo=NULL;
    free(memory.sfix_board);memory.sfix_board=NULL;

    free(memory.pal1);memory.pal1=NULL;
    free(memory.pal2);memory.pal2=NULL;
    free(memory.pal_pc1);memory.pal_pc1=NULL;
    free(memory.pal_pc2);memory.pal_pc2=NULL;
}

void open_bios(void)
{
    FILE *f;
    char *romfile;
    char *path = CF_STR(cf_get_item_by_name("rompath"));//conf.rom_path;
    int len = strlen(path) + 15;

    if (conf.game!=NULL) free_bios_memory();

    /* allocation de la ram */
    memory.ram = (unsigned char *) malloc(0x10000);
    memset(memory.ram,0,0x10000);
    memory.bios = (unsigned char *) malloc(0x20000);
    memory.sfix_board = (unsigned char *) malloc(0x20000);
    memory.ng_lo = (unsigned char *) malloc(0x10000);

    /* partie video */
    memory.pal1 = (unsigned char *) malloc(0x2000);
    memory.pal2 = (unsigned char *) malloc(0x2000);

    memory.pal_pc1 = (unsigned char *) malloc(0x2000);
    memory.pal_pc2 = (unsigned char *) malloc(0x2000);

    memset(memory.video, 0, 0x20000);

    romfile = (char *) malloc(len);
    memset(romfile, 0, len);

    sprintf(romfile, "%s/neo-geo.rom", path);
    f = fopen(romfile, "rb");
    if (f == NULL) {
	printf("Can't find %s\n", romfile);
	exit(1);
    }
    fread(memory.bios, 1, 0x20000, f);
    fclose(f);

    sprintf(romfile, "%s/ng-sfix.rom", path);
    f = fopen(romfile, "rb");
    if (f == NULL) {
	printf("Can't find %s\n", romfile);
	exit(1);
    }
    fread(memory.sfix_board, 1, 0x20000, f);
    fclose(f);

    sprintf(romfile, "%s/ng-lo.rom", path);
    f = fopen(romfile, "rb");
    if (f == NULL) {
	printf("Can't find %s\n", romfile);
	exit(1);
    }
    fread(memory.ng_lo, 1, 0x10000, f);
    fclose(f);


    /* convert bios fix char */
    convert_all_char(memory.sfix_board, 0x20000, memory.fix_board_usage);

    fix_usage = memory.fix_board_usage;
    current_pal = memory.pal1;
    current_fix = memory.sfix_board;
    current_pc_pal = (unsigned short *) memory.pal_pc1;

    free(romfile);
}


